/*=======================================================
  Author:   R. Muth
  Version:  $Id: liveness.c,v 1.56 1999/05/25 23:09:52 muth Exp $
  Comment:  This file contains the primitives for register
            liveness computation and dead code elimination.
=======================================================*/

#include "global.h"
#include "functions.h"

#include "flags.h"
#include "opcode.h"
#include "regs.h"

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "IsDefRange"

extern BOOL IsDefRange(REG reg,INDEX ins1,INDEX ins2)
{
    if( !VALID(ins1) || !VALID(ins2) )
        return 0;

    for(;;)
    {
        if( REG2BIT(reg) & AinsDef(ins1) ) return 1;
        if( ins1 == ins2 ) return 0;
        ins1 = AINS_NEXT(ins1);
    }
    
    return 0;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "RangeUse"

extern BOOL IsUseRange(REG reg,INDEX ins1,INDEX ins2)
{
    if( !VALID(ins1) || !VALID(ins2) )        return 0;
    
    for(;;)
    {
        if( REG2BIT(reg) & AinsUse(ins1) ) return 1;
        if( ins1 == ins2 ) return 0;
        ins1 = AINS_NEXT(ins1);
    }
}

    

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "ComputeSavedChangedRegisters"

extern void ComputeSavedChangedRegisters(FILE *fp)
{
    INDEX caller;
    INDEX callee;
    REGSET current;
    REGSET changed;
    INDEX ifun;
    INDEX iedg;
    BOOL  flag;
    NUMBER count=0;
#ifdef PRINT_INFO
    clock_t time = clock();
#endif

    FOREACH_LIVE_FUN(ifun)
    {
        AfunComputeChangedRegisters(ifun);
        AfunComputeSavedRegisters(ifun);
        AFUN_REGS_CHANGED(ifun) &= ~ AFUN_REGS_SAVED(ifun);
    }
    
    do
    {
        flag = 0;
        count++;
        
        FOREACH_LIVE_EDG_R(iedg)
        {
            if( !AEDG_IS_INTER(iedg) ) continue;

            switch( AEDG_TYPE(iedg) )
            {
            case ET_FUN_CALL:
            case ET_JUMP:
            case ET_NORM:
            case ET_TRUE:
            case ET_FALSE:
            case ET_FUN_LINK:
                caller = ABBL_FUN(AEDG_HEAD(iedg));
                callee = ABBL_FUN(AEDG_TAIL(iedg));
                current = AFUN_REGS_CHANGED(caller);
                changed = AFUN_REGS_CHANGED(callee) & ~AFUN_REGS_SAVED(caller);

                /* if new changed registers found */
                if( (changed | current) != current )
                {
                    flag = 1;
                    AFUN_REGS_CHANGED(caller) = current |changed;
                }
                
                break;
                

            case ET_FUN_RETURN:
            case ET_COMPENSATE:
            case ET_SETLONGJMP:
            case ET_HELL:
                /* do nothing */
                break;
            default:
                ERROR(("Bad interprocedural edge\n"));
                PrintEdgSuper(stdout,iedg);
            }
        }
    }
    while( flag );

#ifdef PRINT_INFO
     time = (clock() - time) * 1000 / CLOCKS_PER_SEC; 
     InfoStatus(fp,"\t[ComputeSavedChanged] Iterations needed: %d {%d ms}\n",count,time);
#endif
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblRegUnmodifiedAfter"

extern BOOL AbblRegUnmodifiedAfter(REG r, INDEX iins)
{
    if (r == REG_ZERO || r == REG_FZERO)
    {
        return TRUE;
    }

    for (iins = AINS_NEXT(iins) ; VALID(iins); iins = AINS_NEXT(iins))
    {
        if (r == AINS_REG_C(iins))
        {
            return FALSE;
        }
    }

    return TRUE;
}

/*=======================================================
@FUN
Registers which are live after an instruction
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsRegsLiveAfter"

extern REGLIST AinsRegsLiveAfter(INDEX iins)
{
    INDEX ii;

    INDEX ibbl = AINS_BBL(iins);
    REGLIST live = ABBL_OUT(ibbl);

    for (ii = ABBL_INS_LAST(ibbl); VALID(ii) && ii != iins; ii = AINS_PREV(ii)) 
    {
        live &= ~AinsDef(ii);
        live |= AinsUse(ii);
    }

    return live;
}


/*=======================================================
@FUN
Registers which are live at the beginning of ibbl
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblRegsLiveIn"

extern REGLIST AbblRegsLiveIn(INDEX ibbl)
{
    INDEX iins;
    REGLIST live;
    INDEX ifun = ABBL_FUN(ibbl);

    if( AFUN_IS_PSEUDO(ifun) )
    {
        switch(ifun)
        {
        case AFUN_PSEUDO_PAL:
            return REGS_PAL_MAY_USE | REGS_PROC_INT_PRESERVED | REGS_PROC_FLT_PRESERVED;
        case AFUN_PSEUDO_CALL:
            return REGS_CALL_MAY_USE | REGS_PROC_INT_PRESERVED | REGS_PROC_FLT_PRESERVED;
        case AFUN_PSEUDO_HELL_LITE:
            /* The really defined registers are only B_V0 and B_A3 but
               others MAY be changed, too. hence we list them all.
               and compensate in AbblRegsUse
               */
	    return B_GP | B_SP | B_RA | REGS_IRET | REGS_FRET | REGS_PROC_INT_PRESERVED | REGS_PROC_FLT_PRESERVED;
        case AFUN_PSEUDO_HELL:
            return REGS_ALL_BUT_ZERO;
            /* The really defined registers are only B_V0 and B_A3 but
               others MAY be changed, too. hence we list them all.
               and compensate in AbblRegsUse
               */
        default:
            ERROR(("unexpected pseudo function: %d [%s]\n",ifun,AFUN_NAME(ifun)));
            return REGS_ALL_BUT_ZERO;
        }
    }
    
    /* since the bbls in AFUN_JSR have no instructions
       the out-sets equal the in-sets */

    live = ABBL_OUT(ibbl);

    FOREACH_BBL_INS_R(ibbl,iins)
    {
        live &= ~AinsDef(iins);
        live |= AinsUse(iins);
    }

    return live;
}
/*=======================================================
@FUN
Registers which Maybe defined in this bbl
it is fairly safe to consider this must information, too!
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblRegsDef"

extern REGLIST AbblRegsDef(INDEX ibbl)
{
    INDEX iins;
    REGLIST def;
    INDEX ifun = ABBL_FUN(ibbl);

    if( AFUN_IS_PSEUDO(ifun) )
    {
        switch(ifun)
        {
        case AFUN_PSEUDO_PAL:
            return REGS_PROC_INT_MAY_CHANGE | REGS_PROC_FLT_MAY_CHANGE;
        case AFUN_PSEUDO_CALL:
            return REGS_PROC_INT_MAY_CHANGE | REGS_PROC_FLT_MAY_CHANGE;
        case AFUN_PSEUDO_HELL_LITE:
            return REGS_PROC_INT_MAY_CHANGE | REGS_PROC_FLT_MAY_CHANGE;
            /* The really defined registers are only B_V0 and B_A3 but
               others MAY be changed, too. hence we list them all.
               and compensate in AbblRegsUse
               */
        case AFUN_PSEUDO_HELL:
            return REGS_ALL_BUT_ZERO;
            /* The really defined registers are only B_V0 and B_A3 but
               others MAY be changed, too. hence we list them all.
               and compensate in AbblRegsUse
               */
        default:
            ERROR(("unexpected pseudo function: %d [%s]\n",ifun,AFUN_NAME(ifun)));
            return REGS_ALL_BUT_ZERO;
        }
    }
    
    def = REGS_NONE;
        
    FOREACH_BBL_INS(ibbl,iins)
    {
        def |= AinsDef(iins);
    }

    return def;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblRegsUse"

extern REGLIST AbblRegsUse(INDEX ibbl)
{
    REGLIST use;
    INDEX iins;
    INDEX ifun = ABBL_FUN(ibbl);

    if( AFUN_IS_PSEUDO(ifun) )
    {
        switch(ifun)
        {

        case AFUN_PSEUDO_PAL:
            return (REGS_PAL_MAY_USE | REGS_PROC_INT_PRESERVED | REGS_PROC_FLT_PRESERVED);
        case AFUN_PSEUDO_CALL:
            return (REGS_CALL_MAY_USE | REGS_PROC_INT_PRESERVED | REGS_PROC_FLT_PRESERVED);
        case AFUN_PSEUDO_HELL_LITE:
            /* hell-lite may clearly use all registers */
	    return B_GP | B_SP | B_RA | REGS_IRET | REGS_FRET | REGS_PROC_INT_PRESERVED | REGS_PROC_FLT_PRESERVED;
        case AFUN_PSEUDO_HELL:
            /* hell may clearly use all registers */
            return REGS_ALL_BUT_ZERO;

        default:
            ERROR(("unexpected pseudo function: %d [%s]\n",ifun,AFUN_NAME(ifun)));
            return REGS_ALL_BUT_ZERO;
        }
    }

    use = REGS_NONE;

    FOREACH_BBL_INS_R(ibbl,iins)
    {
        use &= ~AinsDef(iins);
        use |= AinsUse(iins);
    }

    return use;
}

/*=======================================================
=======================================================*/
/*=======================================================
@COM
This trivial liveness analysis just assumes that every register is live
at the exit of a bbl.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "ComLivenessTrivial"

static void ComLivenessTrivial(FILE *fp)
{
    INDEX ibbl;
        
    /* Initialize the def,use,in,out fields for each bbl */
       
    FOREACH_LIVE_BBL(ibbl)
    {
        ABBL_OUT(ibbl) =  REGS_ALL_BUT_ZERO;
    }

    InfoStatus(fp,"\t[ComLivenessTrivial]\n");
}
/*=======================================================
=======================================================*/
/*=======================================================
@COM
This liveness analysis is the context insensitive one!
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "ComLivenessInsensitive"

static void ComLivenessInsensitive(FILE *fp)
{
    INDEX iedg;
    INDEX ibbl;
    INDEX ifun;
    BOOL  flag;
    NUMBER count=0;
    
    /* Initialize the def,use,in,out fields for each bbl */
       
    FOREACH_LIVE_BBL(ibbl)
    {
        ifun = ABBL_FUN(ibbl);
        if( AFUN_IS_PSEUDO(ifun) )
        {
            /*
              this information is not really used since
              AbblRegLiveIn called below has special handling
              for the PSEUDO functions.
              It will keep these bbls from beeing unecessarily updated, though
              */
            
            ABBL_OUT(ibbl) =  REGS_ALL_BUT_ZERO;
        }
        else
        {
            ABBL_OUT(ibbl) = REGS_NONE;
        }
        
        ABBL_MARK(ibbl);
    }

    /* Do cfa0 fixpoint computation */

    do
    {
        flag = 0;
        count++;
        
        FOREACH_LIVE_BBL_R(ibbl)
        {
            REGLIST in;
            
            if( !ABBL_IS_MARKED(ibbl) )
            {
                continue;
            }
            
            ABBL_UNMARK(ibbl);
            flag = 1;
 
            in = AbblRegsLiveIn(ibbl);

            FOREACH_BBL_PRED(ibbl,iedg)
            {
                INDEX hd = AEDG_HEAD(iedg);
                REGLIST new_out = ABBL_OUT(hd) | in;
                if(  new_out != ABBL_OUT(hd) )
                {
                    ABBL_OUT(hd) = new_out;
                    ABBL_FLAGS(hd) |= F_MARKED;
                }
            }
            
        }
    }
    while( flag );
    
    InfoStatus(fp,"\t[ComLivenessInsensitive] Rounds: %d\n",count);
}

/*=======================================================
=======================================================*/

#define ABBL_REGS_THROUGH(i) ABBL_U1(i).quadword
#define ABBL_REGS_DEF(i) ABBL_U2(i).quadword
#define ABBL_REGS_USE(i) ABBL_U3(i).quadword

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "LivenessFixpoint"

static NUMBER LivenessFixpoint(BOOL allow_edg_ret,int update)
{
    INDEX iedg;
    INDEX ibbl;
    INDEX callee;
    INDEX returnsite;
    
    BOOL  flag;
    NUMBER count=0;
        
    /*
      we assume that bbl and fun stuff has been computed already
      here we just do the do fixpoint computation
      */

    FOREACH_LIVE_BBL(ibbl)
    {
        ABBL_MARK(ibbl);
    }

    do
    {
        flag = 0;
        count++;
        
        FOREACH_LIVE_BBL_R(ibbl)
        {
            REGLIST in,out;

            if( !ABBL_IS_MARKED(ibbl) )
            {
                continue;
            }

            
            
            ABBL_UNMARK(ibbl);
            flag = 1;
 

            in = (ABBL_OUT(ibbl) & ~ABBL_REGS_DEF(ibbl)) | ABBL_REGS_USE(ibbl);
#if 0
            in2 = AbblRegsLiveIn(ibbl);
            if( in != in2 )
            {
                ERROR(("BAD %d   %016lx %016lx   %016lx\n",ibbl,in,in2,in2^in));
            }
#endif            

            if( ABBL_IS_FIRST(ibbl) && !AFUN_IS_PSEUDO(ABBL_FUN(ibbl)) )
            {
#ifdef DO_NOT_USE_SAVE_REGS 
                if( update == 1 )
                    AFUN_REGS_THROUGH(ABBL_FUN(ibbl)) = in;
                else if ( update == 2)
                    AFUN_REGS_USED(ABBL_FUN(ibbl)) = in;
#else 
                if( update == 1 )
                    AFUN_REGS_THROUGH(ABBL_FUN(ibbl)) = in | AFUN_REGS_SAVED(ABBL_FUN(ibbl));
                else if ( update == 2)
                    AFUN_REGS_USED(ABBL_FUN(ibbl)) = in & ~AFUN_REGS_SAVED(ABBL_FUN(ibbl));
#endif
                
            }
            
            FOREACH_BBL_PRED(ibbl,iedg)
            {
                INDEX head = AEDG_HEAD(iedg);
                if( AFUN_IS_PSEUDO(ABBL_FUN(head)) )
                {
                    continue;
                }
                                    
                switch( AEDG_TYPE(iedg) )
                {
                case ET_COMPENSATE:
                case ET_FUN_RETURN:
                    if( allow_edg_ret ) {
                        out = ABBL_OUT(head) | in;
		    }
                    else
                        out = ABBL_OUT(head);
                    break;
                    
                case ET_FUN_CALL:                 
                    callee = ABBL_FUN(ibbl);
                    returnsite = AEDG_TAIL(ABBL_EDG_LINK(head));

                    /* we also need to handle the link edge */
                    
                    out = ABBL_OUT(head) | AFUN_REGS_USED(callee) |
                          (AFUN_REGS_THROUGH(callee) & ((ABBL_OUT(returnsite) & ~ABBL_REGS_DEF(returnsite)) | ABBL_REGS_USE(returnsite)));
                    break;

                case ET_FUN_LINK:
                    callee = ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(head)));
                    out = ABBL_OUT(head) | (AFUN_REGS_THROUGH(callee) & in);
                    break;

                default:
                    out = ABBL_OUT(head) | in;
                    break;
                }
                
                if(  out != ABBL_OUT(head) )
                {
                    ABBL_OUT(head) = out;
                    ABBL_MARK(head);
#if 0
                    if( allow_edg_ret )
                    {
                        ERROR(("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"));
                        PrintEdgSuper(stdout,iedg);
                        
                    }
#endif                    
                }
            }
        }
    }
    while( flag );

    return count;
}

/*=======================================================
=======================================================*/
/*=======================================================
@COM
Context sensitive liveness analysis works as described in the paper
the major pain are the various pseudo functions.
@END
=======================================================*/
#undef __PROC__
#define __PROC__  "ComLivenessSensitive"

/*#define DEBUG_LIVENESS 1 */

static void ComLivenessSensitive(FILE *fp)
{
    INDEX ifun;
    INDEX ibbl;
    
#if 0
    BOOL flag;
#endif
    
    NUMBER rounds1;
    NUMBER rounds2;
    NUMBER rounds3;
    NUMBER rounds4;

    FOREACH_LIVE_FUN(ifun)
    {
        if( AFUN_IS_PSEUDO(ifun) )
        {
            switch(ifun)
            {
            case AFUN_PSEUDO_PAL:
                AFUN_REGS_THROUGH(ifun) = REGS_PROC_INT_PRESERVED | REGS_PROC_FLT_PRESERVED;
                AFUN_REGS_USED(ifun) =  REGS_PAL_MAY_USE;
                break;
            case AFUN_PSEUDO_CALL:
                AFUN_REGS_THROUGH(ifun) = REGS_PROC_INT_PRESERVED | REGS_PROC_FLT_PRESERVED;
                AFUN_REGS_USED(ifun) =  REGS_CALL_MAY_USE;
                break;
	      case AFUN_PSEUDO_HELL_LITE:
                AFUN_REGS_THROUGH(ifun) = REGS_PROC_INT_PRESERVED | REGS_PROC_FLT_PRESERVED;
                AFUN_REGS_USED(ifun) = REGS_CALL_MAY_USE;
                break;
	      case AFUN_PSEUDO_HELL:
                AFUN_REGS_THROUGH(ifun) = REGS_ALL_BUT_ZERO;
                AFUN_REGS_USED(ifun) = REGS_ALL_BUT_ZERO;
                /*                AFUN_REGS_SAVED(ifun) = REGS_NONE; */
                break;
            default:
                ERROR(("unexpected pseudo fun: %d [%s]\n",ifun,AFUN_NAME(ifun)));
                break;
            }
            
        }
        else
        {
            AFUN_REGS_THROUGH(ifun) = REGS_NONE;
            AFUN_REGS_USED(ifun) = REGS_NONE;
        }
    }

    /* this information never changes during the liveness analysis! */

    FOREACH_LIVE_BBL(ibbl)
    {
        ABBL_REGS_USE(ibbl) = AbblRegsUse(ibbl);
        ABBL_REGS_DEF(ibbl) = AbblRegsDef(ibbl);
    }
    
    /*
       initially, we want all exit blocks to have all registers in the out set
       we will propagate this information down into the init blocks
       which then serves as AFUN_REGS_THROUGH
     */
            
    if( ABBL_TYPE(ibbl) == BT_EXIT && !AFUN_IS_PSEUDO(ABBL_FUN(ibbl)) )
    {
        ABBL_OUT(ibbl) =  REGS_ALL_BUT_ZERO;
    }
    else
    {
        ABBL_OUT(ibbl) =  REGS_NONE;
        
    }

    /* do the fixpoint computation */
    
    rounds1 = LivenessFixpoint(0,1);

#ifdef CLEVER_LAST_ROUND
    /* save the through values for clever implementation of the last round */
       
    FOREACH_LIVE_BBL(ibbl)
    {
        ABBL_REGS_THROUGH(ibbl) = ABBL_OUT(ibbl);
    }
#endif

    /*
     * Now we want to compute AFUN_REGS_USED and hence we set the
     * exit blocks to "no register live"
     */

    
    FOREACH_LIVE_BBL(ibbl)
    {
        ABBL_OUT(ibbl) = REGS_NONE;
    }

    rounds2 = LivenessFixpoint(0,2);

    rounds4=0;        

    /*
     * Finally we provide the true information about what is live at the exit block
     * we can use a clever way whoch requires us to save some information earlier
     * or use another fix point iteration.
     */

    
#ifdef CLEVER_LAST_ROUND
    do
    {
        flag = 0;
        rounds4++;
        
        FOREACH_LIVE_FUN(ifun)
        {
            
            INDEX iedg;
            INDEX exit;
            REGLIST out;

            if( AFUN_IS_PSEUDO(ifun) ) continue;
                
            exit = AFUN_BBL_LAST(ifun);
            out=0;
            
            FOREACH_BBL_SUCC(exit,iedg)
            {
                INDEX succ =  AEDG_TAIL(iedg);
                REGLIST in = (ABBL_OUT(succ) & ~ABBL_REGS_DEF(succ)) | ABBL_REGS_USE(succ);
#if 0
                REGLIST in2 = AbblRegsLiveIn(succ);
                ERROR(("BAD \n"));
#endif
                out |= in;
                
            }

            if( out == ABBL_OUT(exit) ) continue;
#if 0
            printf("updating fun %d [%s] %016lx -> %016lx\n",ifun,AFUN_NAME(ifun),ABBL_OUT(exit),out);
#endif            
            if( (out | ABBL_OUT(exit)) != out )
            {
                ERROR(("bad \n"));
                continue;
                
            }
            
            

            
            flag = 1;
            ABBL_OUT(exit) |= out;

            FOREACH_FUN_BBL(ifun,ibbl)
            {
                ABBL_OUT(ibbl) |= (ABBL_REGS_THROUGH(ibbl) & out);
            }
        }
    }
    while( flag );
#else
    rounds3 = LivenessFixpoint(1,0);
#endif
    
    InfoStatus(fp,"\t[ComLivenessSensitive] Rounds: %d %d %d %d\n",rounds1,rounds2,rounds3,rounds4);
}

/*=======================================================
=======================================================*/
/*=======================================================
@COM
The register liveness analysis. This is just a wrapper calling
one of the real liveness analyses depending on the configuration
variable LIVENESS_METHOD which can be set with the command line
switch "-L".
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "ComLiveness"


#define PRINT_INFO_LIVENESS 1

extern void ComLiveness(FILE *fp)
{
    INDEX ibbl;
    INDEX iins;
    NUMBER nb=0;
    NUMBER ni=0;
    NUMBER ib=0;
    NUMBER fb=0;
    NUMBER ii=0;
    NUMBER fi=0;
    REGLIST reglist;

#if PRINT_INFO_LIVENESS    
    clock_t time = clock();
#endif

    if( 0==strcmp("Sensitive",LIVENESS_METHOD) )
    {
        ComLivenessSensitive(fp);
    }
    else if( 0==strcmp("Insensitive",LIVENESS_METHOD) )
    {
        ComLivenessInsensitive(fp);
    }
    else if( 0==strcmp("Trivial",LIVENESS_METHOD) )
    {
        ComLivenessTrivial(fp);
    }
    else
    {
        WARNING(("Bad liveness analysis method %s using Sensitive\n",LIVENESS_METHOD));
        ComLivenessSensitive(fp);
    }

#if PRINT_INFO_LIVENESS    
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC; 
    FOREACH_LIVE_BBL(ibbl)
    {
        reglist =  ABBL_OUT(ibbl);
        ib += BinaryBits( reglist & REGS_INT_BUT_ZERO );
        fb += BinaryBits( reglist & REGS_FLT_BUT_ZERO );
        nb++;

        FOREACH_BBL_INS_R(ibbl,iins)
        {
            ii  += BinaryBits( reglist & REGS_INT_BUT_ZERO );
            fi  += BinaryBits( reglist & REGS_FLT_BUT_ZERO );
            ni++;   
            reglist &= ~AinsDef(iins);
            reglist |= AinsUse(iins);


        }
    }

     InfoStatus(fp,"\t[ComLiveness] INT avg bbl %5.2f  avg ins %5.2f  FLT avg bbl %5.2f  avg ins %5.2f (%d msecs)\n", 
             (double) ib / (double) nb, (double) ii / (double) ni, (double) fb / (double) nb, (double) fi / (double) ni, time);
#endif
}

     

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "OptKillUseless"

extern BOOL OptKillUseless(FILE *fp)
{
    INDEX ibbl;
    INDEX iins;
    REGLIST live;
    REGLIST use;
    REGLIST def;

    INT64  weight = 0;
    NUMBER kcount = 0;
    NUMBER kcounth = 0;

    INT64  lweight = 0;
    
    BOOL change;
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Useless");
    clock_t time = clock();

    FOREACH_LIVE_BBL(ibbl)
    {
        if( AFUN_IS_PSEUDO(ABBL_FUN(ibbl)) ) continue;

        
        /* if( ibbl > PRINT_OPT_NUM ) continue; */
        
        live = ABBL_OUT(ibbl);
    
        FOREACH_BBL_INS_R(ibbl,iins)
        {
            use = AinsUse(iins);
            def = AinsDef(iins);
            
            /*
             * instructions that "write" to REG_ZERO are killed somewhere else
             * this includes FMT_PAL,FMT_IST,FMT_FST,FMT_IBR,FMT_FBR
             */
            
            if( !(def&live) && !AinsHasSideEffect(iins) ) 
            {
                
                if( verbose )
                {
                    printf("\t[*] Killed useless ins in bbl %d:",AINS_BBL(iins));
                    PrintIns(stdout,iins);
                    PrintBblSuper(stdout,ibbl);
                }
                
#if 0
                if( ibbl == PRINT_OPT_NUM )
                {
                    INDEX ibbl = AINS_BBL(iins);
                    INDEX ifun = ABBL_FUN(ibbl);
                    
                    printf(">>>>>>>THIS INS\n");
                    PrintIns(stdout,iins);
                    printf(">>>>>>>THIS BBL\n");
                    PrintBblSuper(stdout,ibbl);
                    /*
                    printf(">>>>>>>NEXT BBL\n");
                    PrintBbl(stdout,ABBL_NEXT(ibbl));;
                    printf(">>>>>>>THIS FUN\n");
                    PrintFunSuper(stdout,ifun);
                    */
                }
#endif
                if( AINS_IS_LOAD(iins) )
                    lweight+= ABBL_COUNT(ibbl);
                weight +=  ABBL_COUNT(ibbl);
                kcount++;
                if( ABBL_IS_HOT(ibbl) )
                    kcounth++;
                AinsKill(iins);


            }
            else
            {
                live &= ~def;
                live |= use;
            }
        }
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    
    change = kcount > 0;

    InfoStatus(fp,"\t[OptUseless%c] Killed Inss: %d#%d lwt: %ld wt %ld {%d ms}\n",
               CHANGE(change),kcount,kcounth,lweight,weight,time);

    return change;
}


/*=======================================================
@OPT

@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptCallingConvention"

extern BOOL OptCallingConvention(FILE *fp)
{
    INDEX ibbl;
    INDEX jbbl;
    INDEX iedg;
    INDEX iins;
    INDEX ifun;
    
    NUMBER ncount=0;
    NUMBER jcount=0;
    NUMBER pcount=0;
    NUMBER fcount=0;
    
    BOOL change;
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Callingconvention");
    clock_t time = clock();

    /* tranform calls to AFUN_PSEUDO_HELL to AFUN_PSEUDO_CALL */

    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_CALL ) continue;
        iins = ABBL_INS_LAST(ibbl);
        if( AINS_CODE(iins) != I_jsr ) continue;

        jcount++;

        if( AINS_REG_C(iins) != REG_RA ) continue;

        if( verbose )
        {
            fprintf(fp,"\t[*]\n");
            PrintBblSuper(stdout,ibbl);
        }
        
        ncount++;
        
        AedgKillFancy( ABBL_EDG_CALL(ibbl) );
        AedgLinkFancy( AedgNewFancy(ibbl,AFUN_BBL_FIRST(AFUN_PSEUDO_CALL),ET_FUN_CALL));
    }


    /* tranform calls to AFUN_PSEUDO_HELL to AFUN_PSEUDO_PAL */

    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_PAL ) continue;

        pcount++;
        
        AedgKillFancy( ABBL_EDG_CALL(ibbl) );
        AedgLinkFancy( AedgNewFancy(ibbl,AFUN_BBL_FIRST(AFUN_PSEUDO_PAL),ET_FUN_CALL));
    }

    

    /*
     * replace calls from hell with calls from pseudo hell which obeys
     * calling conventions
     */    

    jbbl = AFUN_BBL_FIRST(AFUN_PSEUDO_HELL_LITE);

    FOREACH_LIVE_FUN(ifun)
    {
	if (ifun < 4) continue;  /* gross hack: NEEDS WORK */
        
	ibbl = AFUN_BBL_FIRST(ifun);
	if (!VALID(ibbl))
        {
	    continue;
	}

	FOREACH_BBL_PRED(ibbl, iedg)
        {
	    if (AEDG_TYPE(iedg) != ET_HELL) continue;
	    AedgKillFancy(iedg);
	    AedgLinkFancy( AedgNewFancy(jbbl, ibbl, ET_HELL) );
	    fcount++;
	}
    }
                
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    
    change = ncount > 0;

    InfoStatus(fp,"\t[OptCallingconvention%c] pseudo-call %d/%d   pseudo-pal %d   pseudo-hell-lite %d {%d ms}\n",
	       CHANGE(change),ncount,jcount,pcount,fcount,time);

    return change;    
}

    
/*=======================================================
=======================================================*/
/*=======================================================
=======================================================*/


